@@ -42,6 +42,7 @@ from firewall.core.fw_policies import FirewallPolicies
 from firewall.core.fw_ipset import FirewallIPSet
 from firewall.core.fw_transaction import FirewallTransaction
 from firewall.core.fw_helper import FirewallHelper
+from firewall.core.fw_nm import nm_get_bus_name, nm_get_interfaces_in_zone
 from firewall.core.logger import log
 from firewall.core.io.firewalld_conf import firewalld_conf
 from firewall.core.io.direct import Direct
@@ -49,6 +50,7 @@ from firewall.core.io.service import service_reader
 from firewall.core.io.icmptype import icmptype_reader
 from firewall.core.io.zone import zone_reader, Zone
 from firewall.core.io.ipset import ipset_reader
+from firewall.core.ipset import IPSET_TYPES
 from firewall.core.io.helper import helper_reader
 from firewall import errors
 from firewall.errors import FirewallError
@@ -60,24 +62,33 @@ from firewall.errors import FirewallError
 ############################################################################
 
 class Firewall(object):
-    def __init__(self):
+    def __init__(self, offline=False):
         self._firewalld_conf = firewalld_conf(config.FIREWALLD_CONF)
+        self._offline = offline
 
-        self.ip4tables_backend = ipXtables.ip4tables(self)
-        self.ip4tables_enabled = True
-        self.ip4tables_supported_icmp_types = [ ]
-        self.ip6tables_backend = ipXtables.ip6tables(self)
-        self.ip6tables_enabled = True
-        self.ip6tables_supported_icmp_types = [ ]
-        self.ebtables_backend = ebtables.ebtables()
-        self.ebtables_enabled = True
-        self.ipset_backend = ipset.ipset()
-        self.ipset_enabled = True
-        self.ipset_supported_types = [ ]
-        self.nftables_backend = nftables.nftables(self)
-        self.nftables_enabled = True
+        if self._offline:
+            self.ip4tables_enabled = False
+            self.ip6tables_enabled = False
+            self.ebtables_enabled = False
+            self.ipset_enabled = False
+            self.ipset_supported_types = IPSET_TYPES
+            self.nftables_enabled = False
+        else:
+            self.ip4tables_backend = ipXtables.ip4tables(self)
+            self.ip4tables_enabled = True
+            self.ip4tables_supported_icmp_types = [ ]
+            self.ip6tables_backend = ipXtables.ip6tables(self)
+            self.ip6tables_enabled = True
+            self.ip6tables_supported_icmp_types = [ ]
+            self.ebtables_backend = ebtables.ebtables()
+            self.ebtables_enabled = True
+            self.ipset_backend = ipset.ipset()
+            self.ipset_enabled = True
+            self.ipset_supported_types = [ ]
+            self.nftables_backend = nftables.nftables(self)
+            self.nftables_enabled = True
 
-        self.modules_backend = modules.modules()
+            self.modules_backend = modules.modules()
 
         self.icmptype = FirewallIcmpType(self)
         self.service = FirewallService(self)
@@ -91,11 +102,11 @@ class Firewall(object):
         self.__init_vars()
 
     def __repr__(self):
-        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)' % \
+        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)' % \
             (self.__class__, self.ip4tables_enabled, self.ip6tables_enabled,
              self.ebtables_enabled, self._state, self._panic,
              self._default_zone, self._module_refcount, self._marks,
-             self._min_mark, self.cleanup_on_exit, self.ipv6_rpfilter_enabled,
+             self.cleanup_on_exit, self.ipv6_rpfilter_enabled,
              self.ipset_enabled, self._individual_calls, self._log_denied,
              self._automatic_helpers)
 
@@ -106,13 +117,14 @@ class Firewall(object):
         self._module_refcount = { }
         self._marks = [ ]
         # fallback settings will be overloaded by firewalld.conf
-        self._min_mark = config.FALLBACK_MINIMAL_MARK
         self.cleanup_on_exit = config.FALLBACK_CLEANUP_ON_EXIT
         self.ipv6_rpfilter_enabled = config.FALLBACK_IPV6_RPFILTER
         self._individual_calls = config.FALLBACK_INDIVIDUAL_CALLS
         self._log_denied = config.FALLBACK_LOG_DENIED
         self._automatic_helpers = config.FALLBACK_AUTOMATIC_HELPERS
         self._firewall_backend = config.FALLBACK_FIREWALL_BACKEND
+        self._flush_all_on_reload = config.FALLBACK_FLUSH_ALL_ON_RELOAD
+        self._rfc3964_ipv4 = config.FALLBACK_RFC3964_IPV4
         self.nf_conntrack_helper_setting = 0
         self.nf_conntrack_helpers = { }
         self.nf_nat_helpers = { }
@@ -235,9 +247,6 @@ class Firewall(object):
             if self._firewalld_conf.get("DefaultZone"):
                 default_zone = self._firewalld_conf.get("DefaultZone")
 
-            if self._firewalld_conf.get("MinimalMark"):
-                self._min_mark = int(self._firewalld_conf.get("MinimalMark"))
-
             if self._firewalld_conf.get("CleanupOnExit"):
                 value = self._firewalld_conf.get("CleanupOnExit")
                 if value is not None and value.lower() in [ "no", "false" ]:
@@ -298,11 +307,30 @@ class Firewall(object):
                 log.debug1("FirewallBackend is set to '%s'",
                            self._firewall_backend)
 
+            if self._firewalld_conf.get("FlushAllOnReload"):
+                value = self._firewalld_conf.get("FlushAllOnReload")
+                if value.lower() in [ "no", "false" ]:
+                    self._flush_all_on_reload = False
+                else:
+                    self._flush_all_on_reload = True
+                log.debug1("FlushAllOnReload is set to '%s'",
+                           self._flush_all_on_reload)
+
+            if self._firewalld_conf.get("RFC3964_IPv4"):
+                value = self._firewalld_conf.get("RFC3964_IPv4")
+                if value.lower() in [ "no", "false" ]:
+                    self._rfc3964_ipv4 = False
+                else:
+                    self._rfc3964_ipv4 = True
+                log.debug1("RFC3964_IPv4 is set to '%s'",
+                           self._rfc3964_ipv4)
+
         self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))
 
         self._select_firewall_backend(self._firewall_backend)
 
-        self._start_check()
+        if not self._offline:
+            self._start_check()
 
         # load lockdown whitelist
         log.debug1("Loading lockdown whitelist")
@@ -386,6 +414,11 @@ class Firewall(object):
         self.direct.set_permanent_config(obj)
         self.config.set_direct(copy.deepcopy(obj))
 
+        self._default_zone = self.check_zone(default_zone)
+
+        if self._offline:
+            return
+
         # automatic helpers
         #
         # NOTE: must force loading of nf_conntrack to make sure the values are
@@ -447,7 +480,6 @@ class Firewall(object):
         log.debug1("Applying used zones")
         self.zone.apply_zones(use_transaction=transaction)
 
-        self._default_zone = self.check_zone(default_zone)
         self.zone.change_default_zone(None, self._default_zone,
                                       use_transaction=transaction)
 
@@ -646,26 +678,13 @@ class Firewall(object):
         self.__init_vars()
 
     def stop(self):
-        if self.cleanup_on_exit:
+        if self.cleanup_on_exit and not self._offline:
             self.flush()
             self.set_policy("ACCEPT")
             self.modules_backend.unload_firewall_modules()
 
         self.cleanup()
 
-    # marks
-
-    def new_mark(self):
-        # return first unused mark
-        i = self._min_mark
-        while i in self._marks:
-            i += 1
-        self._marks.append(i)
-        return i
-
-    def del_mark(self, mark):
-        self._marks.remove(mark)
-
     # handle modules
 
     def handle_modules(self, _modules, enable):
@@ -798,24 +817,20 @@ class Firewall(object):
 
         if self.is_ipv_enabled("ipv6"):
             ipv6_backend = self.get_backend_by_ipv("ipv6")
-            if self.ipv6_rpfilter_enabled and \
-               "raw" in ipv6_backend.get_available_tables():
+            if "raw" in ipv6_backend.get_available_tables():
+                if self.ipv6_rpfilter_enabled:
+                    rules = ipv6_backend.build_rpfilter_rules(self._log_denied)
+                    transaction.add_rules(ipv6_backend, rules)
 
-                # Execute existing transaction
-                transaction.execute(True)
-                # Start new transaction
-                transaction.clear()
+        if self.is_ipv_enabled("ipv6") and self._rfc3964_ipv4:
+            # Flush due to iptables-restore (nftables) bug triggered when
+            # specifying same index multiple times in same batch
+            # rhbz 1647925
+            transaction.execute(True)
+            transaction.clear()
 
-                rules = ipv6_backend.build_rpfilter_rules(self._log_denied)
-                transaction.add_rules(ipv6_backend, rules)
-
-                # Execute ipv6_rpfilter transaction, it might fail
-                try:
-                    transaction.execute(True)
-                except FirewallError as msg:
-                    log.warning("Applying rules for ipv6_rpfilter failed: %s", msg)
-                # Start new transaction
-                transaction.clear()
+            rules = ipv6_backend.build_rfc3964_ipv4_rules()
+            transaction.add_rules(ipv6_backend, rules)
 
         if use_transaction is None:
             transaction.execute(True)
@@ -962,15 +977,20 @@ class Firewall(object):
     def reload(self, stop=False):
         _panic = self._panic
 
-        # save zone interfaces
-        _zone_interfaces = { }
-        for zone in self.zone.get_zones():
-            _zone_interfaces[zone] = self.zone.get_settings(zone)["interfaces"]
-        # save direct config
-        _direct_config = self.direct.get_runtime_config()
-        _old_dz = self.get_default_zone()
+        # must stash this. The value may change after _start()
+        flush_all = self._flush_all_on_reload
 
-        self.set_policy("DROP")
+        if not flush_all:
+            # save zone interfaces
+            _zone_interfaces = { }
+            for zone in self.zone.get_zones():
+                _zone_interfaces[zone] = self.zone.get_settings(zone)["interfaces"]
+            # save direct config
+            _direct_config = self.direct.get_runtime_config()
+            _old_dz = self.get_default_zone()
+
+        if not _panic:
+            self.set_policy("DROP")
 
         # stop
         self.cleanup()
@@ -983,47 +1003,52 @@ class Firewall(object):
             # etc. We'll re-raise it at the end.
             start_exception = e
 
-        # handle interfaces in the default zone and move them to the new
-        # default zone if it changed
-        _new_dz = self.get_default_zone()
-        if _new_dz != _old_dz:
-            # if_new_dz has been introduced with the reload, we need to add it
-            # https://github.com/firewalld/firewalld/issues/53
-            if _new_dz not in _zone_interfaces:
-                _zone_interfaces[_new_dz] = { }
-            # default zone changed. Move interfaces from old default zone to
-            # the new one.
-            for iface, settings in list(_zone_interfaces[_old_dz].items()):
-                if settings["__default__"]:
-                    # move only those that were added to default zone
-                    # (not those that were added to specific zone same as
-                    # default)
-                    _zone_interfaces[_new_dz][iface] = \
-                        _zone_interfaces[_old_dz][iface]
-                    del _zone_interfaces[_old_dz][iface]
+        if not flush_all:
+            # handle interfaces in the default zone and move them to the new
+            # default zone if it changed
+            _new_dz = self.get_default_zone()
+            if _new_dz != _old_dz:
+                # if_new_dz has been introduced with the reload, we need to add it
+                # https://github.com/firewalld/firewalld/issues/53
+                if _new_dz not in _zone_interfaces:
+                    _zone_interfaces[_new_dz] = { }
+                # default zone changed. Move interfaces from old default zone to
+                # the new one.
+                for iface, settings in list(_zone_interfaces[_old_dz].items()):
+                    if settings["__default__"]:
+                        # move only those that were added to default zone
+                        # (not those that were added to specific zone same as
+                        # default)
+                        _zone_interfaces[_new_dz][iface] = \
+                            _zone_interfaces[_old_dz][iface]
+                        del _zone_interfaces[_old_dz][iface]
 
-        # add interfaces to zones again
-        for zone in self.zone.get_zones():
-            if zone in _zone_interfaces:
-                self.zone.set_settings(zone, { "interfaces":
-                                               _zone_interfaces[zone] })
-                del _zone_interfaces[zone]
-            else:
-                log.info1("New zone '%s'.", zone)
-        if len(_zone_interfaces) > 0:
-            for zone in list(_zone_interfaces.keys()):
-                log.info1("Lost zone '%s', zone interfaces dropped.", zone)
-                del _zone_interfaces[zone]
-        del _zone_interfaces
+            # add interfaces to zones again
+            for zone in self.zone.get_zones():
+                if zone in _zone_interfaces:
+                    self.zone.set_settings(zone, { "interfaces":
+                                                   _zone_interfaces[zone] })
+                    del _zone_interfaces[zone]
+                else:
+                    log.info1("New zone '%s'.", zone)
+            if len(_zone_interfaces) > 0:
+                for zone in list(_zone_interfaces.keys()):
+                    log.info1("Lost zone '%s', zone interfaces dropped.", zone)
+                    del _zone_interfaces[zone]
+            del _zone_interfaces
 
-        # restore direct config
-        self.direct.set_config(_direct_config)
+            # restore direct config
+            self.direct.set_config(_direct_config)
 
-        # enable panic mode again if it has been enabled before or set policy
-        # to ACCEPT
-        if _panic:
-            self.enable_panic_mode()
-        else:
+        # Restore permanent interfaces from NetworkManager
+        nm_bus_name = nm_get_bus_name()
+        if nm_bus_name:
+            for zone in self.zone.get_zones() + [""]:
+                for interface in nm_get_interfaces_in_zone(zone):
+                    self.zone.add_interface(zone, interface, sender=nm_bus_name)
+
+        self._panic = _panic
+        if not self._panic:
             self.set_policy("ACCEPT")
 
         if start_exception:
@@ -1044,9 +1069,8 @@ class Firewall(object):
             raise FirewallError(errors.ALREADY_ENABLED,
                                 "panic mode already enabled")
 
-        # TODO: use rule in raw table not default chain policy
         try:
-            self.set_policy("DROP")
+            self.set_policy("PANIC")
         except Exception as msg:
             raise FirewallError(errors.COMMAND_FAILED, msg)
         self._panic = True
@@ -1056,7 +1080,6 @@ class Firewall(object):
             raise FirewallError(errors.NOT_ENABLED,
                                 "panic mode is not enabled")
 
-        # TODO: use rule in raw table not default chain policy
         try:
             self.set_policy("ACCEPT")
         except Exception as msg:
