@@ -25,7 +25,7 @@ from firewall.core.base import SHORTCUTS, DEFAULT_ZONE_TARGET, \
 from firewall.core.logger import log
 from firewall.functions import portStr, checkIPnMask, checkIP6nMask, \
     checkProtocol, enable_ip_forwarding, check_single_address, check_mac, \
-    portInPortRange
+    portInPortRange, get_nf_conntrack_short_name
 from firewall.core.rich import Rich_Rule, Rich_Accept, \
     Rich_Mark, Rich_Service, Rich_Port, Rich_Protocol, \
     Rich_Masquerade, Rich_ForwardPort, Rich_SourcePort, Rich_IcmpBlock, \
@@ -1588,8 +1588,9 @@ class FirewallZone(object):
                         modules = [ ]
                         for helper in helpers:
                             module = helper.module
+                            _module_short_name = get_nf_conntrack_short_name(module)
                             if self._fw.nf_conntrack_helper_setting == 0:
-                                if helper.name not in \
+                                if _module_short_name not in \
                                    self._fw.nf_conntrack_helpers[module]:
                                     raise FirewallError(
                                         errors.INVALID_HELPER,
@@ -1606,7 +1607,7 @@ class FirewallZone(object):
                                     for (port,proto) in helper.ports:
                                         rules = backend.build_zone_helper_ports_rules(
                                                         enable, zone, proto, port,
-                                                        destination, helper.name)
+                                                        destination, helper.name, _module_short_name)
                                         zone_transaction.add_rules(backend, rules)
                             else:
                                 if helper.module not in modules:
@@ -1803,7 +1804,8 @@ class FirewallZone(object):
             if self._fw.nf_conntrack_helper_setting == 0:
                 for helper in helpers:
                     module = helper.module
-                    if helper.name not in \
+                    _module_short_name = get_nf_conntrack_short_name(module)
+                    if _module_short_name not in \
                        self._fw.nf_conntrack_helpers[module]:
                         raise FirewallError(
                             errors.INVALID_HELPER,
@@ -1820,7 +1822,7 @@ class FirewallZone(object):
                         for (port,proto) in helper.ports:
                             rules = backend.build_zone_helper_ports_rules(
                                             enable, zone, proto, port,
-                                            destination, helper.name)
+                                            destination, helper.name, _module_short_name)
                             zone_transaction.add_rules(backend, rules)
 
             for (port,proto) in svc.ports:
@@ -1877,15 +1879,12 @@ class FirewallZone(object):
             zone_transaction.add_chain("nat", "POSTROUTING")
             zone_transaction.add_chain("filter", "FORWARD_OUT")
 
-        for ipv in ["ipv4", "ipv6"]:
-            zone_transaction.add_post(enable_ip_forwarding, ipv)
+        ipv = "ipv4"
+        zone_transaction.add_post(enable_ip_forwarding, ipv)
 
-        for backend in self._fw.enabled_backends():
-            if not backend.zones_supported:
-                continue
-
-            rules = backend.build_zone_masquerade_rules(enable, zone)
-            zone_transaction.add_rules(backend, rules)
+        backend = self._fw.get_backend_by_ipv(ipv)
+        rules = backend.build_zone_masquerade_rules(enable, zone)
+        zone_transaction.add_rules(backend, rules)
 
     def _forward_port(self, enable, zone, zone_transaction, port, protocol,
                        toport=None, toaddr=None):
