@@ -20,6 +20,7 @@
 #
 
 import os.path
+import copy
 
 from firewall.core.base import SHORTCUTS, DEFAULT_ZONE_TARGET
 from firewall.core.prog import runProg
@@ -27,8 +28,9 @@ from firewall.core.logger import log
 from firewall.functions import tempFile, readfile, splitArgs, check_mac, portStr, \
                                check_single_address
 from firewall import config
-from firewall.errors import FirewallError, INVALID_PASSTHROUGH, INVALID_RULE
-from firewall.core.rich import Rich_Accept, Rich_Reject, Rich_Drop, Rich_Mark
+from firewall.errors import FirewallError, INVALID_PASSTHROUGH, INVALID_RULE, UNKNOWN_ERROR
+from firewall.core.rich import Rich_Accept, Rich_Reject, Rich_Drop, Rich_Mark, \
+                               Rich_Masquerade, Rich_ForwardPort, Rich_IcmpBlock
 import string
 
 BUILT_IN_CHAINS = {
@@ -175,6 +177,8 @@ class ip4tables(object):
         self.restore_wait_option = self._detect_restore_wait_option()
         self.fill_exists()
         self.available_tables = []
+        self.rich_rule_priority_counts = {}
+        self.zone_source_index_cache = []
         self.our_chains = {} # chains created by firewalld
 
     def fill_exists(self):
@@ -286,10 +290,130 @@ class ip4tables(object):
                     chain = args[i+1]
         return (table, chain)
 
+    def _run_replace_zone_source(self, rule, zone_source_index_cache):
+        try:
+            i = rule.index("%%ZONE_SOURCE%%")
+            rule.pop(i)
+            zone = rule.pop(i)
+            if "-m" == rule[4]: # ipset/mac
+                zone_source = (zone, rule[7]) # (zone, address)
+            else:
+                zone_source = (zone, rule[5]) # (zone, address)
+        except ValueError:
+            try:
+                i = rule.index("%%ZONE_INTERFACE%%")
+                rule.pop(i)
+                zone_source = None
+            except ValueError:
+                return
+
+        rule_add = True
+        if rule[0] in ["-D", "--delete"]:
+            rule_add = False
+
+        if zone_source and not rule_add:
+            if zone_source in zone_source_index_cache:
+                zone_source_index_cache.remove(zone_source)
+        elif rule_add:
+            if zone_source:
+                # order source based dispatch by zone name
+                if zone_source not in zone_source_index_cache:
+                    zone_source_index_cache.append(zone_source)
+                    zone_source_index_cache.sort(key=lambda x: x[0])
+
+                index = zone_source_index_cache.index(zone_source)
+            else:
+                index = len(zone_source_index_cache)
+                
+            rule[0] = "-I"
+            rule.insert(2, "%d" % (index + 1))
+
+    def _set_rule_replace_rich_rule_priority(self, rule, rich_rule_priority_counts):
+        """
+        Change something like
+          -t filter -I public_IN %%RICH_RULE_PRIORITY%% 123
+        or
+          -t filter -A public_IN %%RICH_RULE_PRIORITY%% 321
+        into
+          -t filter -I public_IN 4
+        or
+          -t filter -I public_IN
+        """
+        try:
+            i = rule.index("%%RICH_RULE_PRIORITY%%")
+        except ValueError:
+            pass
+        else:
+            rule_add = True
+            insert = False
+            insert_add_index = -1
+            rule.pop(i)
+            priority = rule.pop(i)
+            if type(priority) != int:
+                raise FirewallError(INVALID_RULE, "rich rule priority must be followed by a number")
+
+            table = "filter"
+            for opt in [ "-t", "--table" ]:
+                try:
+                    j = rule.index(opt)
+                except ValueError:
+                    pass
+                else:
+                    if len(rule) >= j+1:
+                        table = rule[j+1]
+            for opt in [ "-A", "--append",
+                         "-I", "--insert",
+                         "-D", "--delete" ]:
+                try:
+                    insert_add_index = rule.index(opt)
+                except ValueError:
+                    pass
+                else:
+                    if len(rule) >= insert_add_index+1:
+                        chain = rule[insert_add_index+1]
+
+                    if opt in [ "-I", "--insert" ]:
+                        insert = True
+                    if opt in [ "-D", "--delete" ]:
+                        rule_add = False
+
+            chain = (table, chain)
+
+            # Add the rule to the priority counts. We don't need to store the
+            # rule, just bump the ref count for the priority value.
+            if not rule_add:
+                if chain not in rich_rule_priority_counts or \
+                   priority not in rich_rule_priority_counts[chain] or \
+                   rich_rule_priority_counts[chain][priority] <= 0:
+                    raise FirewallError(UNKNOWN_ERROR, "nonexistent or underflow of rich rule priority count")
+
+                rich_rule_priority_counts[chain][priority] -= 1
+            else:
+                if chain not in rich_rule_priority_counts:
+                    rich_rule_priority_counts[chain] = {}
+                if priority not in rich_rule_priority_counts[chain]:
+                    rich_rule_priority_counts[chain][priority] = 0
+
+                # calculate index of new rule
+                index = 1
+                for p in sorted(rich_rule_priority_counts[chain].keys()):
+                    if p == priority and insert:
+                        break
+                    index += rich_rule_priority_counts[chain][p]
+                    if p == priority:
+                        break
+
+                rich_rule_priority_counts[chain][priority] += 1
+
+                rule[insert_add_index] = "-I"
+                rule.insert(insert_add_index+2, "%d" % index)
+
     def set_rules(self, rules, log_denied):
         temp_file = tempFile()
 
         table_rules = { }
+        rich_rule_priority_counts = copy.deepcopy(self.rich_rule_priority_counts)
+        zone_source_index_cache = copy.deepcopy(self.zone_source_index_cache)
         for _rule in rules:
             rule = _rule[:]
 
@@ -313,6 +437,9 @@ class ip4tables(object):
                 else:
                     rule.pop(i)
 
+            self._set_rule_replace_rich_rule_priority(rule, rich_rule_priority_counts)
+            self._run_replace_zone_source(rule, zone_source_index_cache)
+
             table = "filter"
             # get table form rule
             for opt in [ "-t", "--table" ]:
@@ -374,6 +501,8 @@ class ip4tables(object):
         if status != 0:
             raise ValueError("'%s %s' failed: %s" % (self._restore_command,
                                                      " ".join(args), ret))
+        self.rich_rule_priority_counts = rich_rule_priority_counts
+        self.zone_source_index_cache = zone_source_index_cache
         return ret
 
     def set_rule(self, rule, log_denied):
@@ -397,7 +526,16 @@ class ip4tables(object):
             else:
                 rule.pop(i)
 
-        return self.__run(rule)
+        rich_rule_priority_counts = copy.deepcopy(self.rich_rule_priority_counts)
+        zone_source_index_cache = copy.deepcopy(self.zone_source_index_cache)
+        self._set_rule_replace_rich_rule_priority(rule, rich_rule_priority_counts)
+        self._run_replace_zone_source(rule, zone_source_index_cache)
+
+        output = self.__run(rule)
+
+        self.rich_rule_priority_counts = rich_rule_priority_counts
+        self.zone_source_index_cache = zone_source_index_cache
+        return output
 
     def get_available_tables(self, table=None):
         ret = []
@@ -447,6 +585,8 @@ class ip4tables(object):
         return wait_option
 
     def build_flush_rules(self):
+        self.rich_rule_priority_counts = {}
+        self.zone_source_index_cache = []
         rules = []
         for table in BUILT_IN_CHAINS.keys():
             if not self.get_available_tables(table):
@@ -460,13 +600,14 @@ class ip4tables(object):
 
     def build_set_policy_rules(self, policy):
         rules = []
+        _policy = "DROP" if policy == "PANIC" else policy
         for table in BUILT_IN_CHAINS.keys():
             if not self.get_available_tables(table):
                 continue
             if table == "nat":
                 continue
             for chain in BUILT_IN_CHAINS[table]:
-                rules.append(["-t", table, "-P", chain, policy])
+                rules.append(["-t", table, "-P", chain, _policy])
         return rules
 
     def supported_icmp_types(self):
@@ -526,11 +667,9 @@ class ip4tables(object):
                 self.our_chains["raw"].add("%s_direct" % chain)
 
                 if chain == "PREROUTING":
-                    default_rules["raw"].append("-N %s_ZONES_SOURCE" % chain)
                     default_rules["raw"].append("-N %s_ZONES" % chain)
-                    default_rules["raw"].append("-A %s -j %s_ZONES_SOURCE" % (chain, chain))
                     default_rules["raw"].append("-A %s -j %s_ZONES" % (chain, chain))
-                    self.our_chains["raw"].update(set(["%s_ZONES_SOURCE" % chain, "%s_ZONES" % chain]))
+                    self.our_chains["raw"].update(set(["%s_ZONES" % chain]))
 
         if self.get_available_tables("mangle"):
             default_rules["mangle"] = [ ]
@@ -541,11 +680,9 @@ class ip4tables(object):
                 self.our_chains["mangle"].add("%s_direct" % chain)
 
                 if chain == "PREROUTING":
-                    default_rules["mangle"].append("-N %s_ZONES_SOURCE" % chain)
                     default_rules["mangle"].append("-N %s_ZONES" % chain)
-                    default_rules["mangle"].append("-A %s -j %s_ZONES_SOURCE" % (chain, chain))
                     default_rules["mangle"].append("-A %s -j %s_ZONES" % (chain, chain))
-                    self.our_chains["mangle"].update(set(["%s_ZONES_SOURCE" % chain, "%s_ZONES" % chain]))
+                    self.our_chains["mangle"].update(set(["%s_ZONES" % chain]))
 
         if self.get_available_tables("nat"):
             default_rules["nat"] = [ ]
@@ -556,21 +693,17 @@ class ip4tables(object):
                 self.our_chains["nat"].add("%s_direct" % chain)
 
                 if chain in [ "PREROUTING", "POSTROUTING" ]:
-                    default_rules["nat"].append("-N %s_ZONES_SOURCE" % chain)
                     default_rules["nat"].append("-N %s_ZONES" % chain)
-                    default_rules["nat"].append("-A %s -j %s_ZONES_SOURCE" % (chain, chain))
                     default_rules["nat"].append("-A %s -j %s_ZONES" % (chain, chain))
-                    self.our_chains["nat"].update(set(["%s_ZONES_SOURCE" % chain, "%s_ZONES" % chain]))
+                    self.our_chains["nat"].update(set(["%s_ZONES" % chain]))
 
         default_rules["filter"] = [
             "-N INPUT_direct",
-            "-N INPUT_ZONES_SOURCE",
             "-N INPUT_ZONES",
 
-            "-A INPUT -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT",
+            "-A INPUT -m conntrack --ctstate RELATED,ESTABLISHED,DNAT -j ACCEPT",
             "-A INPUT -i lo -j ACCEPT",
             "-A INPUT -j INPUT_direct",
-            "-A INPUT -j INPUT_ZONES_SOURCE",
             "-A INPUT -j INPUT_ZONES",
         ]
         if log_denied != "off":
@@ -582,17 +715,13 @@ class ip4tables(object):
 
         default_rules["filter"] += [
             "-N FORWARD_direct",
-            "-N FORWARD_IN_ZONES_SOURCE",
             "-N FORWARD_IN_ZONES",
-            "-N FORWARD_OUT_ZONES_SOURCE",
             "-N FORWARD_OUT_ZONES",
 
-            "-A FORWARD -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT",
+            "-A FORWARD -m conntrack --ctstate RELATED,ESTABLISHED,DNAT -j ACCEPT",
             "-A FORWARD -i lo -j ACCEPT",
             "-A FORWARD -j FORWARD_direct",
-            "-A FORWARD -j FORWARD_IN_ZONES_SOURCE",
             "-A FORWARD -j FORWARD_IN_ZONES",
-            "-A FORWARD -j FORWARD_OUT_ZONES_SOURCE",
             "-A FORWARD -j FORWARD_OUT_ZONES",
         ]
         if log_denied != "off":
@@ -609,10 +738,9 @@ class ip4tables(object):
             "-A OUTPUT -j OUTPUT_direct",
         ]
 
-        self.our_chains["filter"] = set(["INPUT_direct", "INPUT_ZONES_SOURCE", "INPUT_ZONES",
-                                    "FORWARD_direct", "FORWARD_IN_ZONES_SOURCE",
-                                    "FORWARD_IN_ZONES", "FORWARD_OUT_ZONES_SOURCE",
-                                    "FORWARD_OUT_ZONES", "OUTPUT_direct"])
+        self.our_chains["filter"] = set(["INPUT_direct", "INPUT_ZONES",
+                                         "FORWARD_direct", "FORWARD_IN_ZONES",
+                                         "FORWARD_OUT_ZONES", "OUTPUT_direct"])
 
         final_default_rules = []
         for table in default_rules:
@@ -638,9 +766,8 @@ class ip4tables(object):
 
         return {}
 
-    def build_zone_source_interface_rules(self, enable, zone, zone_target,
-                                          interface, table, chain,
-                                          append=False):
+    def build_zone_source_interface_rules(self, enable, zone, interface,
+                                          table, chain, append=False):
         # handle all zones in the same way here, now
         # trust and block zone targets are handled now in __chain
         opt = {
@@ -653,22 +780,22 @@ class ip4tables(object):
         }[chain]
 
         target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS[chain], zone=zone)
-        if zone_target == DEFAULT_ZONE_TARGET:
-            action = "-g"
-        else:
-            action = "-j"
+        action = "-g"
+
         if enable and not append:
-            rule = [ "-I", "%s_ZONES" % chain, "1" ]
+            rule = [ "-I", "%s_ZONES" % chain, "%%ZONE_INTERFACE%%" ]
         elif enable:
             rule = [ "-A", "%s_ZONES" % chain ]
         else:
             rule = [ "-D", "%s_ZONES" % chain ]
+            if not append:
+                rule += ["%%ZONE_INTERFACE%%"]
         rule += [ "-t", table, opt, interface, action, target ]
         return [rule]
 
-    def build_zone_source_address_rules(self, enable, zone, zone_target,
+    def build_zone_source_address_rules(self, enable, zone,
                                         address, table, chain):
-        add_del = { True: "-A", False: "-D" }[enable]
+        add_del = { True: "-I", False: "-D" }[enable]
 
         opt = {
             "PREROUTING": "-s",
@@ -680,10 +807,7 @@ class ip4tables(object):
         }[chain]
 
         target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS[chain], zone=zone)
-        if zone_target == DEFAULT_ZONE_TARGET:
-            action = "-g"
-        else:
-            action = "-j"
+        action = "-g"
 
         if address.startswith("ipset:"):
             name = address[6:]
@@ -693,7 +817,8 @@ class ip4tables(object):
                 opt = "src"
             flags = ",".join([opt] * self._fw.ipset.get_dimension(name))
             rule = [ add_del,
-                     "%s_ZONES_SOURCE" % chain, "-t", table,
+                     "%s_ZONES" % chain, "%%ZONE_SOURCE%%", zone,
+                     "-t", table,
                      "-m", "set", "--match-set", name,
                      flags, action, target ]
         else:
@@ -702,12 +827,14 @@ class ip4tables(object):
                 if opt == "-d":
                     return ""
                 rule = [ add_del,
-                         "%s_ZONES_SOURCE" % chain, "-t", table,
+                         "%s_ZONES" % chain, "%%ZONE_SOURCE%%", zone,
+                         "-t", table,
                          "-m", "mac", "--mac-source", address.upper(),
                          action, target ]
             else:
                 rule = [ add_del,
-                         "%s_ZONES_SOURCE" % chain, "-t", table,
+                         "%s_ZONES" % chain, "%%ZONE_SOURCE%%", zone,
+                         "-t", table,
                          opt, address, action, target ]
         return [rule]
 
@@ -717,16 +844,22 @@ class ip4tables(object):
         self.our_chains[table].update(set([_zone,
                                       "%s_log" % _zone,
                                       "%s_deny" % _zone,
+                                      "%s_pre" % _zone,
+                                      "%s_post" % _zone,
                                       "%s_allow" % _zone]))
 
         rules = []
         rules.append([ "-N", _zone, "-t", table ])
+        rules.append([ "-N", "%s_pre" % _zone, "-t", table ])
         rules.append([ "-N", "%s_log" % _zone, "-t", table ])
         rules.append([ "-N", "%s_deny" % _zone, "-t", table ])
         rules.append([ "-N", "%s_allow" % _zone, "-t", table ])
+        rules.append([ "-N", "%s_post" % _zone, "-t", table ])
+        rules.append([ "-A", _zone, "-t", table, "-j", "%s_pre" % _zone ])
         rules.append([ "-A", _zone, "-t", table, "-j", "%s_log" % _zone ])
         rules.append([ "-A", _zone, "-t", table, "-j", "%s_deny" % _zone ])
         rules.append([ "-A", _zone, "-t", table, "-j", "%s_allow" % _zone ])
+        rules.append([ "-A", _zone, "-t", table, "-j", "%s_post" % _zone ])
 
         target = self._fw.zone._zones[zone].target
 
@@ -760,13 +893,53 @@ class ip4tables(object):
             return [ "-m", "limit", "--limit", limit.value ]
         return []
 
+    def _rich_rule_chain_suffix(self, rich_rule):
+        if type(rich_rule.element) in [Rich_Masquerade, Rich_ForwardPort, Rich_IcmpBlock]:
+            # These are special and don't have an explicit action
+            pass
+        elif rich_rule.action:
+            if type(rich_rule.action) not in [Rich_Accept, Rich_Reject, Rich_Drop, Rich_Mark]:
+                raise FirewallError(INVALID_RULE, "Unknown action %s" % type(rich_rule.action))
+        else:
+            raise FirewallError(INVALID_RULE, "No rule action specified.")
+
+        if rich_rule.priority == 0:
+            if type(rich_rule.element) in [Rich_Masquerade, Rich_ForwardPort] or \
+               type(rich_rule.action) in [Rich_Accept, Rich_Mark]:
+                return "allow"
+            elif type(rich_rule.element) in [Rich_IcmpBlock] or \
+                 type(rich_rule.action) in [Rich_Reject, Rich_Drop]:
+                return "deny"
+        elif rich_rule.priority < 0:
+            return "pre"
+        else:
+            return "post"
+
+    def _rich_rule_chain_suffix_from_log(self, rich_rule):
+        if not rich_rule.log and not rich_rule.audit:
+            raise FirewallError(INVALID_RULE, "Not log or audit")
+
+        if rich_rule.priority == 0:
+            return "log"
+        elif rich_rule.priority < 0:
+            return "pre"
+        else:
+            return "post"
+
+    def _rich_rule_priority_fragment(self, rich_rule):
+        if rich_rule.priority == 0:
+            return []
+        return ["%%RICH_RULE_PRIORITY%%", rich_rule.priority]
+
     def _rich_rule_log(self, rich_rule, enable, table, target, rule_fragment):
         if not rich_rule.log:
             return []
 
         add_del = { True: "-A", False: "-D" }[enable]
 
-        rule = [ add_del, "%s_log" % (target), "-t", table]
+        chain_suffix = self._rich_rule_chain_suffix_from_log(rich_rule)
+        rule = ["-t", table, add_del, "%s_%s" % (target, chain_suffix)]
+        rule += self._rich_rule_priority_fragment(rich_rule)
         rule += rule_fragment + [ "-j", "LOG" ]
         if rich_rule.log.prefix:
             rule += [ "--log-prefix", "'%s'" % rich_rule.log.prefix ]
@@ -782,7 +955,10 @@ class ip4tables(object):
 
         add_del = { True: "-A", False: "-D" }[enable]
 
-        rule = [add_del, "%s_log" % (target), "-t", table] + rule_fragment
+        chain_suffix = self._rich_rule_chain_suffix_from_log(rich_rule)
+        rule = ["-t", table, add_del, "%s_%s" % (target, chain_suffix)]
+        rule += self._rich_rule_priority_fragment(rich_rule)
+        rule += rule_fragment
         if type(rich_rule.action) == Rich_Accept:
             _type = "accept"
         elif type(rich_rule.action) == Rich_Reject:
@@ -802,28 +978,28 @@ class ip4tables(object):
 
         add_del = { True: "-A", False: "-D" }[enable]
 
+        chain_suffix = self._rich_rule_chain_suffix(rich_rule)
+        chain = "%s_%s" % (target, chain_suffix)
         if type(rich_rule.action) == Rich_Accept:
-            chain = "%s_allow" % target
             rule_action = [ "-j", "ACCEPT" ]
         elif type(rich_rule.action) == Rich_Reject:
-            chain = "%s_deny" % target
             rule_action = [ "-j", "REJECT" ]
             if rich_rule.action.type:
                 rule_action += [ "--reject-with", rich_rule.action.type ]
         elif type(rich_rule.action) ==  Rich_Drop:
-            chain = "%s_deny" % target
             rule_action = [ "-j", "DROP" ]
         elif type(rich_rule.action) == Rich_Mark:
             target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS["PREROUTING"],
                                                 zone=zone)
             table = "mangle"
-            chain = "%s_allow" % target
+            chain = "%s_%s" % (target, chain_suffix)
             rule_action = [ "-j", "MARK", "--set-xmark", rich_rule.action.set ]
         else:
             raise FirewallError(INVALID_RULE,
                                 "Unknown action %s" % type(rich_rule.action))
 
-        rule = [ add_del, chain, "-t", table ]
+        rule = ["-t", table, add_del, chain]
+        rule += self._rich_rule_priority_fragment(rich_rule)
         rule += rule_fragment + rule_action
         rule += self._rule_limit(rich_rule.action.limit)
 
@@ -945,7 +1121,7 @@ class ip4tables(object):
         return rules
 
     def build_zone_helper_ports_rules(self, enable, zone, proto, port,
-                                      destination, helper_name):
+                                      destination, helper_name, module_short_name):
         add_del = { True: "-A", False: "-D" }[enable]
         target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS["PREROUTING"],
                                             zone=zone)
@@ -954,7 +1130,7 @@ class ip4tables(object):
             rule += [ "--dport", "%s" % portStr(port) ]
         if destination:
             rule += [ "-d",  destination ]
-        rule += [ "-j", "CT", "--helper", helper_name ]
+        rule += [ "-j", "CT", "--helper", module_short_name ]
 
         return [rule]
 
@@ -964,11 +1140,15 @@ class ip4tables(object):
                                             zone=zone)
         rule_fragment = []
         if rich_rule:
+            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
+            rule_fragment += self._rich_rule_priority_fragment(rich_rule)
             rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
             rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
+        else:
+            chain_suffix = "allow"
 
         rules = []
-        rules.append([ add_del, "%s_allow" % (target), "-t", "nat" ]
+        rules.append(["-t", "nat", add_del, "%s_%s" % (target, chain_suffix)]
                      + rule_fragment +
                      [ "!", "-o", "lo", "-j", "MASQUERADE" ])
         # FORWARD_OUT
@@ -976,22 +1156,23 @@ class ip4tables(object):
                                             zone=zone)
         rule_fragment = []
         if rich_rule:
+            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
+            rule_fragment += self._rich_rule_priority_fragment(rich_rule)
             rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
             rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
+        else:
+            chain_suffix = "allow"
 
-        rules.append([ add_del, "%s_allow" % (target), "-t", "filter"]
+        rules.append(["-t", "filter", add_del, "%s_%s" % (target, chain_suffix)]
                      + rule_fragment +
                      ["-m", "conntrack", "--ctstate", "NEW,UNTRACKED", "-j", "ACCEPT" ])
 
         return rules
 
-    def build_zone_forward_port_rules(self, enable, zone, filter_chain, port,
-                                      protocol, toport, toaddr, mark_id, rich_rule=None):
+    def build_zone_forward_port_rules(self, enable, zone, port,
+                                      protocol, toport, toaddr, rich_rule=None):
         add_del = { True: "-A", False: "-D" }[enable]
 
-        mark_str = "0x%x" % mark_id
-        mark = [ "-m", "mark", "--mark", mark_str ]
-
         to = ""
         if toaddr:
             if check_single_address("ipv6", toaddr):
@@ -1004,29 +1185,22 @@ class ip4tables(object):
         target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS["PREROUTING"],
                                             zone=zone)
 
-        rule_fragment = [ "-p", protocol, "--dport", portStr(port) ]
+        rule_fragment = []
         if rich_rule:
+            chain_suffix = self._rich_rule_chain_suffix(rich_rule)
+            rule_fragment = self._rich_rule_priority_fragment(rich_rule)
             rule_fragment += self._rich_rule_destination_fragment(rich_rule.destination)
             rule_fragment += self._rich_rule_source_fragment(rich_rule.source)
+        else:
+            chain_suffix = "allow"
 
         rules = []
         if rich_rule:
-            rules.append(self._rich_rule_log(rich_rule, enable, "mangle", target, rule_fragment))
-        rules.append([ add_del, "%s_allow" % (target), "-t", "mangle"]
-                     + rule_fragment + 
-                     [ "-j", "MARK", "--set-mark", mark_str ])
-
-        # local and remote
-        rules.append([ add_del, "%s_allow" % (target), "-t", "nat",
-                     "-p", protocol ] + mark +
-                     [ "-j", "DNAT", "--to-destination", to ])
-
-        target = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS[filter_chain],
-                                            zone=zone)
-        rules.append([ add_del, "%s_allow" % (target),
-                     "-t", "filter", "-m", "conntrack",
-                     "--ctstate", "NEW,UNTRACKED" ] +
-                     mark + [ "-j", "ACCEPT" ])
+            rules.append(self._rich_rule_log(rich_rule, enable, "nat", target, rule_fragment))
+        rules.append(["-t", "nat", add_del, "%s_%s" % (target, chain_suffix)]
+                     + rule_fragment +
+                     ["-p", protocol, "--dport", portStr(port),
+                      "-j", "DNAT", "--to-destination", to])
 
         return rules
 
@@ -1064,7 +1238,9 @@ class ip4tables(object):
                 if rich_rule.action:
                     rules.append(self._rich_rule_action(zone, rich_rule, enable, table, target, rule_fragment))
                 else:
-                    rules.append([ add_del, "%s_deny" % target, "-t", table ]
+                    chain_suffix = self._rich_rule_chain_suffix(rich_rule)
+                    rules.append(["-t", table, add_del, "%s_%s" % (target, chain_suffix)]
+                                 + self._rich_rule_priority_fragment(rich_rule)
                                  + rule_fragment +
                                  [ "-j", "%%REJECT%%" ])
             else:
@@ -1083,7 +1259,7 @@ class ip4tables(object):
         table = "filter"
         rules = []
         for chain in [ "INPUT", "FORWARD_IN" ]:
-            rule_idx = 4
+            rule_idx = 6
             _zone = DEFAULT_ZONE_TARGET.format(chain=SHORTCUTS[chain],
                                                zone=zone)
 
@@ -1155,3 +1331,37 @@ class ip6tables(ip4tables):
                        "--icmpv6-type=router-advertisement",
                        "-j", "ACCEPT" ]) # RHBZ#1058505
         return rules
+
+    def build_rfc3964_ipv4_rules(self):
+        daddr_list = [
+                     "::0.0.0.0/96", # IPv4 compatible
+                     "::ffff:0.0.0.0/96", # IPv4 mapped
+                     "2002:0000::/24", # 0.0.0.0/8 (the system has no address assigned yet)
+                     "2002:0a00::/24", # 10.0.0.0/8 (private)
+                     "2002:7f00::/24", # 127.0.0.0/8 (loopback)
+                     "2002:ac10::/28", # 172.16.0.0/12 (private)
+                     "2002:c0a8::/32", # 192.168.0.0/16 (private)
+                     "2002:a9fe::/32", # 169.254.0.0/16 (IANA Assigned DHCP link-local)
+                     "2002:e000::/19", # 224.0.0.0/4 (multicast), 240.0.0.0/4 (reserved and broadcast)
+                     ]
+
+        chain_name = "RFC3964_IPv4"
+        self.our_chains["filter"].add(chain_name)
+
+        rules = []
+        rules.append(["-t", "filter", "-N", chain_name])
+        for daddr in daddr_list:
+            rules.append(["-t", "filter", "-I", chain_name,
+                          "-d", daddr, "-j", "REJECT", "--reject-with",
+                          "addr-unreach"])
+            if self._fw._log_denied in ["unicast", "all"]:
+                rules.append(["-t", "filter", "-I", chain_name,
+                              "-d", daddr, "-j", "LOG",
+                              "--log-prefix", "\"RFC3964_IPv4_REJECT: \""])
+
+        # Inject into FORWARD and OUTPUT chains
+        rules.append(["-t", "filter", "-I", "OUTPUT", "3",
+                      "-j", chain_name])
+        rules.append(["-t", "filter", "-I", "FORWARD", "4",
+                      "-j", chain_name])
+        return rules
